@@ -14,14 +14,16 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 import errno
+import logging
 import os
-import weakref
 
 import tooz
 from tooz import coordination
 from tooz.drivers import _retry
 from tooz import locking
 
+LOG = logging.getLogger(__name__)
+
 
 class FileLock(locking.Lock):
     """A file based lock."""
@@ -29,9 +31,9 @@ class FileLock(locking.Lock):
     def __init__(self, path):
         super(FileLock, self).__init__(path)
         self.acquired = False
+        self.lockfile = open(self.name, 'a')
 
     def acquire(self, blocking=True):
-        self.lockfile = open(self.name, 'a')
 
         @_retry.retry(stop_max_delay=blocking)
         def _lock():
@@ -57,7 +59,6 @@ class FileLock(locking.Lock):
 
     def release(self):
         self.unlock()
-        self.lockfile.close()
         self.acquired = False
 
     def lock(self):
@@ -66,6 +67,10 @@ class FileLock(locking.Lock):
     def unlock(self):
         raise NotImplementedError
 
+    def __del__(self):
+        if self.acquired:
+            LOG.warn("unreleased lock %s garbage collected" % self.name)
+
 
 class WindowsFileLock(FileLock):
     def lock(self):
@@ -94,8 +99,6 @@ else:
 class FileDriver(coordination.CoordinationDriver):
     """A file based driver."""
 
-    LOCKS = weakref.WeakValueDictionary()
-
     def __init__(self, member_id, parsed_url, options):
         """Initialize the file driver."""
         super(FileDriver, self).__init__()
@@ -103,8 +106,7 @@ class FileDriver(coordination.CoordinationDriver):
 
     def get_lock(self, name):
         path = os.path.abspath(os.path.join(self._lockdir, name.decode()))
-        lock = LockClass(path)
-        return self.LOCKS.setdefault(path, lock)
+        return locking.SharedWeakLockHelper(self._lockdir, LockClass, path)
 
     @staticmethod
     def watch_join_group(group_id, callback):
