@@ -20,8 +20,13 @@ import contextlib
 from distutils import version
 import logging
 
+try:
+    from time import monotonic as _now  # noqa
+except ImportError:
+    from time import time as _now  # noqa
+
 from concurrent import futures
-from oslo.utils import strutils
+from oslo_utils import strutils
 import redis
 from redis import exceptions
 from redis import lock as redis_locks
@@ -343,13 +348,14 @@ class RedisDriver(coordination.CoordinationDriver):
             sentinel_server = sentinel.Sentinel(
                 sentinel_hosts,
                 socket_timeout=kwargs['socket_timeout'])
-            master_host, master_port = sentinel_server.discover_master(
-                kwargs['sentinel'])
-            kwargs['host'] = master_host
-            kwargs['port'] = master_port
+            sentinel_name = kwargs['sentinel']
             del kwargs['sentinel']
             if 'sentinel_fallback' in kwargs:
                 del kwargs['sentinel_fallback']
+            master_client = sentinel_server.master_for(sentinel_name, **kwargs)
+            # The master_client is a redis.StrictRedis using a
+            # Sentinel managed connection pool.
+            return master_client
         return redis.StrictRedis(**kwargs)
 
     def _start(self):
@@ -664,10 +670,17 @@ class RedisDriver(coordination.CoordinationDriver):
         raise tooz.NotImplemented
 
     def run_watchers(self, timeout=None):
+        if timeout is not None:
+            started_at = _now()
         result = []
         for group_id in self.get_groups().get(timeout=timeout):
+            leftover_timeout = None
+            if timeout is not None:
+                elapsed = max(0.0, _now() - started_at)
+                leftover_timeout = timeout - elapsed
             try:
-                group_members = self.get_members(group_id).get(timeout=timeout)
+                group_members = self.get_members(group_id).get(
+                    timeout=leftover_timeout)
             except coordination.GroupNotCreated:
                 group_members = set()
             else:
