@@ -19,10 +19,15 @@ import copy
 
 from kazoo import client
 from kazoo import exceptions
+from kazoo.handlers import eventlet as eventlet_handler
+from kazoo.handlers import threading as threading_handler
 from kazoo.protocol import paths
+from oslo_utils import strutils
 import six
+from six.moves import filter as compat_filter
 
 from tooz import coordination
+from tooz.drivers import _retry
 from tooz import locking
 from tooz import utils
 
@@ -31,13 +36,34 @@ class ZooKeeperLock(locking.Lock):
     def __init__(self, name, lock):
         super(ZooKeeperLock, self).__init__(name)
         self._lock = lock
+        self.acquired = False
 
     def acquire(self, blocking=True):
-        return self._lock.acquire(blocking=bool(blocking),
-                                  timeout=blocking)
+
+        @_retry.retry(stop_max_delay=blocking)
+        def _lock():
+            # NOTE(sileht): kazoo lock looks broken:
+            # https://github.com/python-zk/kazoo/issues/291
+            # so we track the exclusivity internally
+            if self.acquired is True:
+                if blocking:
+                    raise _retry.Retry
+                return False
+
+            if self._lock.acquire(blocking=bool(blocking),
+                                  timeout=0):
+                self.acquired = True
+                return True
+
+            if blocking:
+                raise _retry.Retry
+            return False
+
+        return _lock()
 
     def release(self):
-        return self._lock.release()
+        self._lock.release()
+        self.acquired = False
 
 
 class BaseZooKeeperDriver(coordination.CoordinationDriver):
@@ -51,6 +77,8 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 
     def __init__(self, member_id, parsed_url, options):
         super(BaseZooKeeperDriver, self).__init__()
+        options = utils.collapse(options, exclude=['hosts'])
+        self._options = options
         self._member_id = member_id
         self.timeout = int(options.get('timeout', ['10'])[-1])
 
@@ -72,6 +100,14 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
     def _stop(self):
         self._coord.stop()
 
+    @staticmethod
+    def _dumps(data):
+        return utils.dumps(data)
+
+    @staticmethod
+    def _loads(blob):
+        return utils.loads(blob)
+
     @staticmethod
     def _create_group_handler(async_result, timeout,
                               timeout_exception, group_id):
@@ -93,6 +129,27 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
                               timeout_exception=self._timeout_exception,
                               group_id=group_id)
 
+    @staticmethod
+    def _delete_group_handler(async_result, timeout,
+                              timeout_exception, group_id):
+        try:
+            async_result.get(block=True, timeout=timeout)
+        except timeout_exception as e:
+            raise coordination.OperationTimedOut(utils.exception_message(e))
+        except exceptions.NoNodeError:
+            raise coordination.GroupNotCreated(group_id)
+        except exceptions.NotEmptyError:
+            raise coordination.GroupNotEmpty(group_id)
+        except exceptions.ZookeeperError as e:
+            raise coordination.ToozError(utils.exception_message(e))
+
+    def delete_group(self, group_id):
+        group_path = self._path_group(group_id)
+        async_result = self._coord.delete_async(group_path)
+        return ZooAsyncResult(async_result, self._delete_group_handler,
+                              timeout_exception=self._timeout_exception,
+                              group_id=group_id)
+
     @staticmethod
     def _join_group_handler(async_result, timeout,
                             timeout_exception, group_id, member_id):
@@ -109,6 +166,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 
     def join_group(self, group_id, capabilities=b""):
         member_path = self._path_member(group_id, self._member_id)
+        capabilities = self._dumps(capabilities)
         async_result = self._coord.create_async(member_path,
                                                 value=capabilities,
                                                 ephemeral=True)
@@ -170,13 +228,14 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 
     def update_capabilities(self, group_id, capabilities):
         member_path = self._path_member(group_id, self._member_id)
+        capabilities = self._dumps(capabilities)
         async_result = self._coord.set_async(member_path, capabilities)
         return ZooAsyncResult(async_result, self._update_capabilities_handler,
                               timeout_exception=self._timeout_exception,
                               group_id=group_id, member_id=self._member_id)
 
-    @staticmethod
-    def _get_member_capabilities_handler(async_result, timeout,
+    @classmethod
+    def _get_member_capabilities_handler(cls, async_result, timeout,
                                          timeout_exception, group_id,
                                          member_id):
         try:
@@ -188,7 +247,7 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         except exceptions.ZookeeperError as e:
             raise coordination.ToozError(utils.exception_message(e))
         else:
-            return capabilities
+            return cls._loads(capabilities)
 
     def get_member_capabilities(self, group_id, member_id):
         member_path = self._path_member(group_id, member_id)
@@ -238,15 +297,49 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
 class KazooDriver(BaseZooKeeperDriver):
     """The driver using the Kazoo client against real ZooKeeper servers."""
 
+    HANDLERS = {
+        'eventlet': eventlet_handler.SequentialEventletHandler,
+        'threading': threading_handler.SequentialThreadingHandler,
+    }
+    """
+    Restricted immutable dict of handler 'kinds' -> handler classes that
+    this driver can accept via 'handler' option key (the expected value for
+    this option is one of the keys in this dictionary).
+    """
+
     def __init__(self, member_id, parsed_url, options):
         super(KazooDriver, self).__init__(member_id, parsed_url, options)
-        self._coord = self._make_client(parsed_url, options)
+        self._coord = self._make_client(parsed_url, self._options)
         self._member_id = member_id
         self._timeout_exception = self._coord.handler.timeout_exception
 
-    @classmethod
-    def _make_client(cls, parsed_url, options):
-        return client.KazooClient(hosts=parsed_url.netloc)
+    def _make_client(self, parsed_url, options):
+        # Creates a kazoo client,
+        # See: https://github.com/python-zk/kazoo/blob/1.3.1/kazoo/client.py
+        # for what options a client takes...
+        maybe_hosts = [parsed_url.netloc] + list(options.get('hosts', []))
+        hosts = list(compat_filter(None, maybe_hosts))
+        if not hosts:
+            hosts = ['localhost:2181']
+        randomize_hosts = options.get('randomize_hosts', True)
+        client_kwargs = {
+            'hosts': ",".join(hosts),
+            'timeout': float(options.get('timeout', self.timeout)),
+            'connection_retry': options.get('connection_retry'),
+            'command_retry': options.get('command_retry'),
+            'randomize_hosts': strutils.bool_from_string(randomize_hosts),
+        }
+        handler_kind = options.get('handler')
+        if handler_kind:
+            try:
+                handler_cls = self.HANDLERS[handler_kind]
+            except KeyError:
+                raise ValueError("Unknown handler '%s' requested"
+                                 " valid handlers are %s"
+                                 % (handler_kind,
+                                    sorted(self.HANDLERS.keys())))
+            client_kwargs['handler'] = handler_cls()
+        return client.KazooClient(**client_kwargs)
 
     def _watch_group(self, group_id):
         get_members_req = self.get_members(group_id)
